from .git import Git
from .base import Local
from .zenodo import Zenodo
from .figshare import Figshare
from .dataverse import Dataverse
from .dataworkspace import Dataworkspace
from .hydroshare import Hydroshare
from .mercurial import Mercurial
